// $Id: Connection.cc,v 1.31.2.10 1998/11/25 18:16:23 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include "Connection.h"
#include "Clock.h"
#include "LocalStats.h"
#include "ErrHist.h"
#include "PortMgr.h"
#include "Log.h"
#include "Gadgets.h"

static class ConnStats {
	struct ConnStatsRec {
		ConnStatsRec(): thePendingLevel(0), thePumpingLevel(0), theCcLevel(0) {}
		int thePendingLevel;
		int thePumpingLevel;
		int theCcLevel;
	};
	public:
		inline void update(int dPend, int dPump);

	protected:
		ConnStatsRec theLevels;
		ConnStatsRec theRecords;
		Count theCount;
} TheConnStats;


inline
void ConnStats::update(int dPend, int dPump) {

	if ((theLevels.thePendingLevel += dPend) > theRecords.thePendingLevel)
		theRecords.thePendingLevel = theLevels.thePendingLevel;

	if ((theLevels.thePumpingLevel += dPump) > theRecords.thePumpingLevel)
		theRecords.thePumpingLevel = theLevels.thePumpingLevel;

	theLevels.theCcLevel = theLevels.thePumpingLevel + theLevels.thePendingLevel;
	if (theLevels.theCcLevel > theRecords.theCcLevel)
		theRecords.theCcLevel = theLevels.theCcLevel;

	if (0 && theCount++ % 50000 == 0)
		SysLog << "ConnStats: "
			<< theLevels.thePendingLevel << " <= " << theRecords.thePendingLevel
			<< " / "
			<< theLevels.thePumpingLevel << " <= " << theRecords.thePumpingLevel
			<< " : " << theLevels.theCcLevel << " <= " << theRecords.theCcLevel
			<< ende;
}

bool Connection::open() {
	assert(theAddr.host() && theAddr.port());
	if (!theSocket.create()) {
		TheErrHist[errno]++;
		SysLog << "connection " << this << " failed to create a socket; cause(?): " << strerror(errno) << ende;
		return false;
	}

	//theSocket.reuseAddr(true);
	//theSocket.reusePort(true);
	theSocket.blocking(false);
	theSocket.linger(0);

	Journal << "connection " << this << " fd: " << (int)theSocket << " ~opened to " << theAddr << ende;
	// bind to an ephemeral port then connect
	theLocPort = ThePortMgr.bind(theSocket);
	const bool res = theLocPort >= 0 && theSocket.connect(theAddr.host(), theAddr.port());
	if (res) {
		StatInc(theConnConnectCount);
		TheConnStats.update(+1, 0);
		theState = csPending;
		theStart = TheClock;
		Journal << "connected to " << theSocket.rport() << endl;
	} else {
		TheErrHist[errno]++;
		theSocket.close(); // close here
		if (theLocPort >= 0)
			ThePortMgr.free(theLocPort, false);
	}
	return res;
}

bool Connection::accept(Socket listenSock, bool &fatal) {
	struct sockaddr_in addr;
	int addr_len = sizeof(addr);
	theSocket = listenSock.accept((struct sockaddr*)&addr, &addr_len);
	if (theSocket >= 0) {
		theAddr.host(inet_ntoa(addr.sin_addr));
		theAddr.port(ntohs(addr.sin_port));
		StatInc(theConnAcceptCount);
		theIOReadyFlags[ioRead] = false; theIOReadyFlags[ioWrite] = true;
		TheConnStats.update(+1, 0);
		theState = csPending;
		theStart = TheClock;
		Journal << "connection " << this << " fd: " << (int)theSocket << " accepted from " << theAddr << ende;
	} else {
		fatal = errno != EWOULDBLOCK;
		if (fatal) {
			TheErrHist[errno]++;
			SysLog << "connection " << this << " failed to accept from " << listenSock << " cause(?): " << strerror(errno) << ende;
		}
		Notepad << "connection " << this << " failed to accept from " << listenSock << " cause(?): " << strerror(errno) << ende;
	}
	return theSocket >= 0;
}

void Connection::close() {
	Journal << "connection " << this << " fd: " << (int)theSocket << " closed to " << theAddr << ende;
	theIOReadyFlags[ioRead] = theIOReadyFlags[ioWrite] = theIOReadyFlags[ioExcept] = false;
	theSocket.close();
	if (theLocPort >= 0)
		ThePortMgr.free(theLocPort, !theReaderBuf.err() && !theWriterBuf.err());
	StatInc(theConnCloseCount);
	TheConnStats.update(
		theState == csPending ? -1 : 0,
		theState == csPumping ? -1 : 0);
	theState = csClosed;
}

// from the socket into the rBuf
Size Connection::read(Size max_size) {
	theReaderBuf.stretch(max_size);
	assert(theIOReadyFlags[ioRead]); // remove if flags are not guaranteed to be set
	theIOReadyFlags[ioRead] = false;

	if (theState == csPending) {
		const int e = theSocket.errNo();
		if (e) {
			SysLog << "connection " << this << " connect error: " << e << ' ' << strerror(e) << ende;
			SysLog << " read:  " << theReaderBuf.inOffset() << '/' << theReaderBuf.outOffset() << ende;
			SysLog << " wrote: " << theWriterBuf.outOffset() << '/' << theWriterBuf.inOffset() << ende;
			SysLog << " delay: " << TheClock.now() - theStart << ende;
			theReaderBuf.err(e);
			theReaderBuf.eod(true);
			TheErrHist[e]++;
			return -1;
		}
		TheConnStats.update(-1, +1);
		theState = csPumping;
	}

	const ssize_t sz = ::read(theSocket, theReaderBuf.buf(), max_size);
	theReaderBuf.err(sz < 0 ? errno : 0);

	Journal << "connection " << this << " fd: " << (int)theSocket << " read " << sz << " bytes (max: " << max_size << ")" << ende;
	theReaderBuf.eod(sz <= 0);
	if (sz > 0)
		theReaderBuf.appended(sz);
	if (sz < 0) {
		TheErrHist[errno]++;
		SysLog << "connection " << this << " read error: " << theReaderBuf.errNo() << ' ' << strerror(theReaderBuf.errNo()) << ende;
		SysLog << " read:  " << theReaderBuf.inOffset() << '/' << theReaderBuf.outOffset() << ende;
		SysLog << " wrote: " << theWriterBuf.outOffset() << '/' << theWriterBuf.inOffset() << ende;
		SysLog << " delay: " << TheClock.now() - theStart << ende;
	}
	StatInc(theConnReadCount);

	// KLUDGE
	static int readCount = 10;
	if (readCount-- > 0)
		cout << "# read[" << readCount << "] " << sz << " bytes from " << theSocket.peerAddr() << ende;

	return sz;
}

// from the wBuf into the socket
Size Connection::write(Size max_size) {

	if (theState == csPending) {
		const int e = theSocket.errNo();
		if (e) {
			SysLog << "connection " << this << " accept error: " << e << ' ' << strerror(e) << ende;
			SysLog << " read:  " << theReaderBuf.inOffset() << '/' << theReaderBuf.outOffset() << ende;
			SysLog << " wrote: " << theWriterBuf.outOffset() << '/' << theWriterBuf.inOffset() << ende;
			SysLog << " delay: " << TheClock.now() - theStart << ende;
			theReaderBuf.err(e);
			theReaderBuf.eod(true);
			TheErrHist[e]++;
			return -1;
		}
		TheConnStats.update(-1, +1);
		theState = csPumping;
	}

	if (max_size <= 0)
		return 0;
	assert(max_size <= theWriterBuf.contSize());
	assert(theIOReadyFlags[ioWrite]); // remove if flags are not guaranteed to be set
	theIOReadyFlags[ioWrite] = false;

	const ssize_t sz = ::write(theSocket, theWriterBuf.content(), max_size);
	theWriterBuf.err(sz < 0 ? errno : 0);

	Journal << "connection " << this << " fd: " << (int)theSocket << " wrote " << sz << " bytes (max: " << max_size << ")" << ende;
	if (sz > 0)
		theWriterBuf.consumed(sz);
	if (sz < 0) {
		TheErrHist[errno]++;
		SysLog << "connection " << this << " write error: " << theWriterBuf.errNo() << ' ' << strerror(theWriterBuf.errNo()) << ende;
	}
	StatInc(theConnWriteCount);

	// KLUDGE
	static int writeCount = 10;
	if (writeCount-- > 0)
		cout << "# wrote[" << writeCount << "] " << sz << " bytes to " << theSocket.peerAddr() << ende;

	return sz;
}
