// $Id: SelectLoop.cc,v 1.10.2.5 1998/11/16 18:14:48 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h" /* to reset FD_SETSIZE, in part */

#include <unistd.h>
#include <errno.h>

#include "SelectLoop.h"
#include "Clock.h"
#include "LocalStats.h"
#include "ErrHist.h"
#include "Log.h"
#include "Gadgets.h"

SelectLoop TheSelectLoop;

static IOer::IOMethod IOMethods[3] = {
	&IOer::readReady,
	&IOer::writeReady,
	&IOer::exceptReady
};
static const char *IOMethodNames[3] = {
	"readReady",
	"writeReady",
	"exceptReady"
};


FD_Set::FD_Set(): theIOers(0), theMaxFD(-1), theResCount(0) {

	FD_ZERO(&theSet);
	FD_ZERO(&theReadySet);

	theIOers = new (IOer *[FD_SETSIZE]);
	memset(theIOers, 0, sizeof(IOer*)*FD_SETSIZE);
}

FD_Set::~FD_Set() {
	delete[] theIOers;
}

void FD_Set::setFD(int fd, IOer *p) {

	assert(fd >= 0);
	assert(fd < FD_SETSIZE);

	if (p) { // set
		assert(!theIOers[fd]);
		FD_SET(fd, &theSet);
		if (fd > theMaxFD)
			theMaxFD = fd;
		theResCount++;
	} else { // clear
		assert(fd <= theMaxFD);
		assert(theIOers[fd]);
		FD_CLR(fd, &theSet);
		FD_CLR(fd, &theReadySet); // just in case
		if (fd == theMaxFD)
			while (--theMaxFD >= 0 && !FD_ISSET(theMaxFD, &theSet));
		theResCount--;
	}
	/* will clear or set */
	theIOers[fd] = p;
}

// call IOers for each ready file
void FD_Set::doIO(int fd, IOer::IOMethod method, const char *methodName) {
	IOer *p = theIOers[fd];
	assert(p);
	Notepad << "Select: " << p << "->" << methodName << '('  << fd << ')' << ende;
	(p->*method)(fd);
}

#if UNUSED_CODE
void FD_Set::report(ostream &os) const {

	os << "registered fd:" << endl;

	for (int fd = 0; fd <= theMaxFD; ++fd) {
		if (FD_ISSET(fd, &theReadySet) || FD_ISSET(fd, &theSet))
			os << setw(5) << fd
				<< (FD_ISSET(fd, &theReadySet) ? " r " : "   ")
				<< (FD_ISSET(fd, &theSet) ? " s " : "   ") 
				<< endl;
	}
}
#endif

SelectLoop::SelectLoop(): theMaxFD(-1), theResCount(0) {
}


// runs one select loop cycle
// returns #ready fds and -1 on error
int SelectLoop::step(Time timeout, int tickCount = 0) {
	Assert(tickCount <= 0 || theTicker);
	Journal 
		<< "call select(" << theMaxFD << ',' << timeout << ")"
		<< " clients: " << theResCount << ende;

	const Time start = TheClock.now();
	const int readyCount = select(theMaxFD+1,
		theSets[0].makeReadySet(),
		theSets[1].makeReadySet(),
		theSets[2].makeReadySet(),
		timeout >= 0 ? &timeout : 0);
	const Time delay = TheClock.now() - start;

	Journal << "select() took " << delay << " usec; ready: " << readyCount << '/' << theResCount << ende;

	if (readyCount > 0) {
		const int tickDelta = tickCount > 0 ? readyCount/(1+tickCount) : 0;
		int t = 0;
		for (int fd = theMaxFD; fd >= 0; --fd) {
			// ignore exceptions
			// XXX: move Methods to Set class
			if (theSets[0].isReady(fd)) {
				theSets[0].doIO(fd, IOMethods[0], IOMethodNames[0]);
				t++;
			}
			if (theSets[1].isReady(fd)) {
				theSets[1].doIO(fd, IOMethods[1], IOMethodNames[1]);
				t++;
			}
			if (t && tickCount > 0 && t >= tickDelta) {
				(*theTicker)();
				t = 0;
				tickCount--;
			}
		}
	} else
	if (readyCount < 0 && errno != EINTR) {
		TheErrHist[errno]++;
		SysLog << "select failed, cause(?): " << strerror(errno) << ende;
		abort();
	}

	return readyCount;
}

SelectReserv SelectLoop::setFD(int fd, IODir dir, IOer *p) {
	assert(p);
	theSets[dir].setFD(fd, p);
	if (fd >= theMaxFD)
		theMaxFD = fd;
	theResCount++;
	Notepad << p << " waits for " << IOMethodNames[dir] << '(' << fd << ") clients: " << theResCount << ende;
	return SelectReserv(fd, dir);
}

void SelectLoop::clearFD(int fd, IODir dir) {
	assert(fd >= 0 && dir != ioNone);
	theResCount--;
	Notepad << theSets[dir][fd] << " clears " << IOMethodNames[dir] << '(' << fd << ") clients: " << theResCount << ende;
	theSets[dir].setFD(fd, 0);
	if (fd >= theMaxFD)
		syncMaxFD();
}

// synchronizes global maxFD with those in the sets
void SelectLoop::syncMaxFD() {
	theMaxFD = max(theSets[0].maxFD(),
		max(theSets[1].maxFD(), theSets[2].maxFD()));
}
