#ifndef _MUTEX_CLASS_H_
#define _MUTEX_CLASS_H_ "$Id: Mutex.h,v 1.8.2.1 1998/09/22 07:44:14 rousskov Exp $"

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"
#include "Queue.h"
#include "Identity.h"
#include "Log.h"



class ResMgr;
class Mutexer;

// a Mutex protects a resource that cannot be shared (mutually exclusive locking)
class Mutex {
	public:
		Mutex(): theId(-1) {}
		Mutex(ResMgr *aMgr, int anId): theMgr(aMgr), theId(anId) {}
		~Mutex() {}

		operator bool() const { return locked(); }
		bool queued() const { return theId >= 0; }
		bool operator ==(const Mutex &mx) const { return mx.theMgr == theMgr && mx.theId == theId && theId >= 0; }

		void obtain(Mutexer *mxer);
		void releaseValid();
		void releaseVoid();
		void giveUpWaiting(Mutexer *mxer);

	protected:
		inline bool locked() const;

	protected:
		ResMgr *theMgr;
		int theId;
};

// owner of the resource being mutexed
class ResOwner: public virtual Identity {
	public:
		typedef void (ResOwner::*IMethod)(ResMgr *);

	public:
		virtual ~ResOwner() {}

		virtual void idleRes(ResMgr *) {}
		virtual void voidRes(ResMgr *) {}
};

// a user of the resource being mutexed (mutex holder)
class Mutexer: public virtual Identity {
	public:
		typedef void (Mutexer::*IMethod)(const Mutex &mx);

	public:
		virtual ~Mutexer() {}

		virtual void obtainedLock(const Mutex &mx) = 0; // got what was asking for
		virtual void deprivedLock(const Mutex &mx) = 0; // lost obtained lock
		virtual void refusedLock(const Mutex &mx) = 0;  // will never obtain
};

// resource manager
// - maintains a queue of users waiting for a resource
//   (first guy in the queue is the current user of the resource)
class ResMgr {
	protected:
		typedef Queue<Mutexer*> MxQueue;
		typedef void (ResMgr::*IMethod)(Mutexer *mxer);

	public:
		ResMgr(ResOwner *anOwner, int aResId = -1, int aResIdx = -1): theOwner(anOwner), theResId(aResId), theResIdx(aResIdx), theMxId(0), theLastMxId(0) {}
		~ResMgr() { dequeueAll(&depriveMx, &refuseMx); }

		int resId() const { return theResId; }
		int resIdx() const { return theResIdx; }
		bool idle() const { return theQueue.empty(); }
		bool locked(int mxId) const { return mxId == theMxId && !theQueue.empty(); }
		MxQueue *queue() { return &theQueue; }

		void resId(int anId) { theResId = anId; }
		void resIdx(int anIdx) { theResIdx = anIdx; }

		// get a mutex, no reservation is made (but you must get one to make it)
		Mutex getMx() { return Mutex(this, ++theLastMxId); }

		void obtain(Mutexer *mxer) { theQueue.enqueue(mxer); if (theQueue.length() == 1) giveMx(mxer); }
		void releaseValid(int mxId) { Assert(locked(mxId)); shift(&giveMx); }
		void releaseVoid(int mxId) { Assert(locked(mxId)); dequeueAll(0, &refuseMx); inform(&ResOwner::voidRes, "voidRes"); }
		void forget(Mutexer *mxer);

	protected:
		void shift(IMethod im);
		void dequeueAll(IMethod im1, IMethod im);

		void newMx(Mutexer::IMethod mim, const char *name, Mutexer *mxer) { theMxId++; inform(mim, name, mxer); }
		void giveMx(Mutexer *mxer) { newMx(&Mutexer::obtainedLock, "obtainedLock", mxer); }
		void depriveMx(Mutexer *mxer) { newMx(&Mutexer::deprivedLock, "deprivedLock", mxer); }
		void refuseMx(Mutexer *mxer) { newMx(&Mutexer::refusedLock, "refusedLock", mxer); }
		
		inline void inform(Mutexer::IMethod im, const char *iname, Mutexer *mxer);
		inline void inform(ResOwner::IMethod im, const char *iname);

	protected:
		MxQueue theQueue;
		ResOwner *theOwner;
		int theResId;
		int theResIdx;
		int theMxId;
		int theLastMxId;
};



/* inline methods */

inline
void Mutex::obtain(Mutexer *mxer) {
	Assert(theMgr);
	theMgr->obtain(mxer);
}

inline
void Mutex::releaseValid() {
	Assert(theMgr);
	theMgr->releaseValid(theId);
}

inline
void Mutex::releaseVoid() {
	Assert(theMgr);
	theMgr->releaseVoid(theId);
}

inline
void Mutex::giveUpWaiting(Mutexer *mxer) {
	Assert(theMgr);
	theMgr->forget(mxer);
}

inline
bool Mutex::locked() const {
	return theMgr && theMgr->locked(theId);
}

inline
void ResMgr::inform(Mutexer::IMethod im, const char *name, Mutexer *mxer) {
	Notepad << mxer << "->" << name << '(' << theMxId << ')' << ende;
	(mxer->*im)(Mutex(this, theMxId));
}

inline
void ResMgr::inform(ResOwner::IMethod im, const char *name) {
	Notepad << theOwner << "->" << name << '(' << this << ')' << ende;
	(theOwner->*im)(this);
}


#endif
