// $Id: main.cc,v 1.49.2.19 1998/11/25 18:16:27 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include <stdlib.h>
#include <signal.h>
#include <new.h>

#include <fstream.h>
#include <iomanip.h>

#include "Size.h"
#include "Array.h"
#include "Client.h"
#include "Server.h"
#include "AlarmClock.h"
#include "SelectLoop.h"
#include "ConnMgr.h"
#include "LoadGen.h"
#include "ConstAddrGen.h"
#include "ConstAddrRndReqMaker.h"
#include "RndRepMaker.h"
#include "RepEater.h"
#include "ReqEater.h"
#include "BlockReader.h"
#include "BlockWriter.h"
#include "ProgressMeter.h"
#include "LocalStats.h"
#include "ErrHist.h"
#include "PortMgr.h"
#include "PhaseMgr.h"
#include "Log.h"
#include "Gadgets.h"
#include "globals.h"


// local vars

// true if we should exit main loop nomatterwhat
static bool TheShutDownFlag = false;

static Array<Client*> clients;
static Array<Server*> servers;

// import
extern RateMeter TheSubmMeter;
extern RateMeter TheComplMeter;
extern RateMeter TheErrMeter;

// saves the values of real meters at the end of the measurement phase
static PhaseMgr ThePhaseMgr;


static 
void SelTicker() {
	static int i = 0;
	if (int c = servers.count())
		servers[i++ % c]->checkPorts();
}

// main loop: iterate until no alarms are set and select loops is idle
static 
void loop() {

 	while (!TheAlarmClock.isClear() || !TheSelectLoop.isIdle() || !TheKernel.idle()) {
		UpdateClocks();

		//const int lastComplCount = ThePhaseStats->theComplMeter.count();
		while (!TheKernel.idle()) {
			TheKernel.slice();
			UpdateClocks();
		}
		static const int complDelta = servers.count() ? 10 : 0; // ThePhaseStats->theComplMeter.count() - lastComplCount;

		// if alarm is not set, use the default delay
		Time time_left = TheAlarmClock.isClear() ? 
			TheOpt.theIdleTimeout : TheAlarmClock.timeLeft();

		// call select even if the SelectLoop is idle (but there are alarms)
		if (!TheSelectLoop.isIdle() || !TheAlarmClock.isClear()) {
			const int count = TheSelectLoop.step(time_left, complDelta);
			// stop if busy select returned nothing after max delay
			if (!count && !TheSelectLoop.isIdle() && TheAlarmClock.isClear() && time_left >= TheOpt.theIdleTimeout) {
				TheShutDownFlag = true;
				SysLog << "# shutting down after " << time_left << " of inactivity" << ende;
			}
		}

		ThePhaseMgr.checkPhase();

		// break if shut down requested
		if (TheShutDownFlag)
			break;
	}
}

static
void outOfMem() {
	cerr << "out of memory in new" << endl;
	abort();
}

static
void report(int id, const RateMeter &subm, const RateMeter &compl, const RateMeter &err) {
	if (id >= 0)
		Report << setw(6) << id;
	else
		Report << (id == -1 ? "middle" : "sum");

	Report
		<< "\t " << subm.count()
		<< "\t " << subm
		<< "\t " << compl.count()
		<< "\t " << compl
		<< "\t " << err.count()
		<< "\t " << (subm.count() - compl.count())
		<< endl;
}

template <class TopProc>
static
void report(const char *prefix, const Array<TopProc*> &procs) {
	Report << endl;
	if (!procs.count()) {
		Report << "# " << prefix << ": none" << endl;
		return;
	}

	// header
	Report << prefix
		<< "\t subm"
		<< "\t s/sec"
		<< "\t compl"
		<< "\t c/sec"
		<< "\t errs"
		<< "\t left"
		<< endl;
	// report for all processes
	RateMeter submM, complM, errM;
	for (int i = 0; i < procs.count(); ++i) {
		complM += procs[i]->complMeter();
		submM += procs[i]->submMeter();
		errM += procs[i]->errMeter();
		procs[i]->report(Report << setw(6) << i << "\t ");
		Report << endl;
	}
	// totals
	report(-2, submM, complM, errM);
}

static
void shutDown(int) {
	TheShutDownFlag = true;
	signal(SIGINT, SIG_DFL);
}

static
void configure() {

	// set resource limits
	TheMaxFD = SetMaxFD(TheMaxFD);
	assert(TheMaxFD > 0);

	ThePortMgr.configure(TheOpt.theMinPort, TheOpt.theMaxPort,
		TheOpt.useEphPorts());

	if (TheOpt.theServerCount)
		TheSelectLoop.ticker(&SelTicker);

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);

	// quit nicely on SIGINT
	signal(SIGINT, &shutDown);

	// custom out-of-memory handler
	set_new_handler(&outOfMem);

	// log format
	ConfigureLogs(2);

	// world params
	TheWorld.id(TheOpt.theWorldId);
	TheWorld.capacity(TheOpt.theWorldCapacity);
	TheWorld.objSizeMean(TheOpt.theObjSizeMean);
	if (TheOpt.theUrlsFileName && !TheWorld.loadUrls(TheOpt.theUrlsFileName)) {
		SysLog << "failed to load urls from '" << TheOpt.theUrlsFileName << "'" << endl;
		exit(-3);
	}
	if (TheOpt.theReqTotal > 0)
		TheWorld.goal(TheOpt.theReqTotal);
	else
	if (TheOpt.theClientCount > 0)
		TheWorld.goal(TheOpt.theClientCount*TheOpt.theReqPerClient);
}


static
int usage(const char *prg_name) {
	static char *Id = StrDup("Polygraph. Version 0.0p2. $Date: 1998/11/25 18:16:27 $ (c) 1998");

	for (char *p = Id; *p; ++p)
		if (*p != '$') cerr << *p;
	cerr << endl;

	cerr << "usage: " << prg_name
		<< " [--option ...]"
		<< " client_specs"
		<< " [listen_port# ...]"
		<< endl;

	cerr << "client_specs: <num><'x'|'t'><count>['r'<rate>]" << endl
		<< "  <num>                 number of concurrent clients" << endl
		<< "                        (0 for server-only mode)" << endl
		<< "  x<count>              requests per client (unlimited by default)" << endl
		<< "  t<count>              requests total" << endl
		<< "  r<rate>               'Poisson' workload: cumulative req rate" << endl
		<< "                          rate is an average submission rate in req/sec" << endl
		<< "  requests are submitted as replies are received" << endl
		<< "  ('stream' workload) unless <rate> is specified" << endl;

	cerr << "options:" << endl
		<< "  --host <name>         origin server host (affects URL generation)" << endl
		<< "  --proxy <IP:port#>    proxy address (affects connect(2) requests)" << endl
		<< "  --world <id:count>    numerical world id and #objects in the world" << endl
		<< "                          use 'max' for maximum #objects" << endl
		<< "  --world <filename>    preload URLs from file; regular URL format used" << endl
		<< "                          for objects with ids larger than #loaded urls" << endl
		<< "  --seq <seed>          sequential object ids; ids start with seed" << endl
		<< "  --rnd <seed>          random object ids; r.n.g seeded with seed" << endl
		<< "                          ids are distributed as zipf(1)" << endl
		<< "  --unique <seed>       append unique ids to urls; ids start with seed" << endl
		<< "  --obj_size <kbytes>   mean object size" << endl
		<< "  --serv_think <sec>    server think time (per write)" << endl
		<< "  --rep_cachable <0|1>  generate [non]cachable replies" << endl
		<< "  --launch_win <sec>    window to start all clients" << endl
		<< "  --ports <min:max>     explicit port range for connect requests" << endl
		<< "  --ports eph           use ephemeral ports for connect requests" << endl
		<< "  --idle_tout <sec>     exit after <sec> of inactivity" << endl
		<< "  --idle_tout none      wait for some activity forever" << endl
		<< "  --conn_est_tout <msec> connection establishment timeout" << endl
		<< "                          support is experimental" << endl;

	cerr << "generated URL format (abridged):" << endl
		<< "  http://host_name/obj_id/world_id" << endl
		<< " or" << endl
		<< "  http://host_name/obj_id/world_id/unique_id" << endl;

	return -1;
}

int main(int argc, char *argv[]) {

	if (!TheOpt.parse(argc, argv))
		return usage(argv[0]);

	configure();

	UpdateClocks();
	Time startTime = TheClock;

	Journal << ThePrgName << " started." << ende;
	Report << "Command:";
	for (int i = 0; i < argc; ++i)
		Report << ' ' << argv[i];
	Report << endl;

	TheOpt.report(Report);
	Report << "Max_FD: " << TheMaxFD << endl;

	ThePhaseMgr.start();

	// proxy address (affects client connect requests)
	const Address &addrProxy = TheOpt.theProxyAddr;

	// default server configuration
	ServerCfg scfg;
	scfg.theConnMgr = new ConnMgr(0);
	scfg.thePause = new Pause();
	scfg.theXactCfg.theContMaker = new RndRepMaker(Size::KB(4), false);
	scfg.theXactCfg.theReader = new BlockReader(Size::KB(4));
	scfg.theXactCfg.theWriter = new BlockWriter(Size::KB(4));
	scfg.theXactCfg.theContEater = new ReqEater();

	const int client_count = TheOpt.theClientCount;
	const int server_count = TheOpt.theServerCount;
	const IntArray &ports = TheOpt.theServerPorts;

	// configuration for the client's connection mgr
	ConnMgrCfg cConnMgrCfg;
	cConnMgrCfg.theMaxCcUseLevel = 1; /* no cc requests on one connection */
	cConnMgrCfg.theMaxTotUseCount = 1; /* no persistent connections at all */

	// progress meter configuration for clients
	ProgMeterCfg pmCfg;
	pmCfg.theMaxCcLevel = 16; // max cc requests at a time
	pmCfg.theMaxStartCount = TheOpt.theReqPerClient;

	// loader configuration for clients
	Loader *loader = 0;
	if (TheOpt.theClientThinkTime >= 0) { // LoadGen ("real")
		LoadGenCfg cfgLoadGen;
		cfgLoadGen.theBlobSize = 1;
		cfgLoadGen.theBlobCount = -1; /* no limit */
		cfgLoadGen.theIntraBlobPause = new ConstPause(Time::Sec(0));
		cfgLoadGen.theInterBlobPause =
			new RndPause(new ExpDistr(new RndGen(), TheOpt.theClientThinkTime.secd()));
		LoadGen *lg = new LoadGen();
		lg->configure(cfgLoadGen);
		loader = lg;
	} else {
		loader = new StreamLoader(1);
	}
	assert(loader);

	// default client configuration
	ClientCfg ccfg;
	ccfg.theConnMgr = new ConnMgr(0);
	ccfg.theConnMgr->configure(cConnMgrCfg);
	ccfg.theProgressMeter = new ProgressMeter();
	ccfg.theProgressMeter->configure(pmCfg);
	ccfg.theLoader = loader;
	ccfg.theAddrGen = new ConstAddrGen(addrProxy);
	ccfg.theXactCfg.theContMaker = new ConstAddrRndReqMaker(TheOpt.theOriginAddr);
	ccfg.theXactCfg.theReader = new BlockReader(Size::KB(4));
	ccfg.theXactCfg.theWriter = new BlockWriter(Size::KB(4));
	ccfg.theXactCfg.theContEater = new RepEater();

	// create and start some servers
	for (int i = 0; i < server_count; ++i) {
		Server *s = new Server();
		s->configure(scfg.clone());
		s->port(ports[i]);
		servers.append(s);
		TheKernel.start(s);
	}

	// create and start some clients
	for (int i = 0; i < client_count; ++i) {
		Client *c = new Client();
		c->configure(ccfg.clone());
		clients.append(c);
		TheKernel.start(clients[i]);
	}

	loop();

	UpdateClocks(false);
	Journal << "all jobs completed in " << (TheClock-startTime) << ende;
	Report << "execution time: " << (TheClock-startTime) << endl;

	// collect final reports
	TheErrHist.report(Report << endl);
	Report << endl;
	ThePhaseMgr.lastReport();
	report("server", servers);
	report("client", clients);

	// cleanup
	{
		for (int i = 0; i < clients.count(); ++i)
			delete clients[i];
		for (int i = 0; i < servers.count(); ++i) {
			TheKernel.stop(servers[i]);
			delete servers[i];
		}
		scfg.destroy();
		ccfg.destroy();
	}

	Journal << ThePrgName << " done." << ende;

	return 0;
}
